@@ -1,7 +1,6 @@
 /*
     SPDX-FileCopyrightText: 2020 David Edmundson <davidedmundson@kde.org>
     SPDX-FileCopyrightText: 2021 Méven Car <meven.car@enioka.com>
-
     SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
@@ -13,7 +12,10 @@
 #include <QImageReader>
 #include <QImageWriter>
 #include <QMimeData>
+#include <QMutex>
 #include <QPointer>
+#include <QThread>
+#include <QTimer>
 #include <QWaylandClientExtension>
 #include <QWindow>
 #include <QtWaylandClientVersion>
@@ -28,6 +30,14 @@
 #include "qwayland-wayland.h"
 #include "qwayland-ext-data-control-v1.h"
 
+/**
+ * Wayland clipboard wraps ext_data_control an additional is used as we need to avoid
+ * deadlocks if an application tries to read the normal clipboard whilst it owns the data control.
+ *
+ * To solve this all ext_data_control classes live on another thread which dispatches events on a separate
+ * queue. A recursive mutex allows the main thread to read mimedata and no wayland events which change the mimedata process until this is complete.
+ */
+
 static inline QString applicationQtXImageLiteral()
 {
     return QStringLiteral("application/x-qt-image");
@@ -298,10 +308,6 @@ public:
     {
         return m_mimeData.get();
     }
-    std::unique_ptr<QMimeData> releaseMimeData()
-    {
-        return std::move(m_mimeData);
-    }
 
 Q_SIGNALS:
     void cancelled();
@@ -322,7 +328,7 @@ DataControlSource::DataControlSource(struct ::ext_data_control_source_v1 *id, QM
     for (const QString &format : formats) {
         offer(format);
     }
-    if (mimeData->hasText()) {
+    if (mimeData->hasText() && !m_mimeData->hasFormat(utf8Text())) {
         // ensure GTK applications get this mimetype to avoid them discarding the offer
         offer(QStringLiteral("text/plain;charset=utf-8"));
     }
@@ -340,7 +346,7 @@ DataControlSource::DataControlSource(struct ::ext_data_control_source_v1 *id, QM
 void DataControlSource::ext_data_control_source_v1_send(const QString &mime_type, int32_t fd)
 {
     QString send_mime_type = mime_type;
-    if (send_mime_type == QStringLiteral("text/plain;charset=utf-8")) {
+    if (send_mime_type == utf8Text() && !m_mimeData->hasFormat(utf8Text())) {
         // if we get a request on the fallback mime, send the data from the original mime type
         send_mime_type = QStringLiteral("text/plain");
     }
@@ -445,6 +451,7 @@ protected:
 
     void ext_data_control_device_v1_selection(struct ::ext_data_control_offer_v1 *id) override
     {
+        QMutexLocker locker(&m_selectionLock);
         if (!id) {
             m_receivedSelection.reset();
         } else {
@@ -457,6 +464,7 @@ protected:
 
     void ext_data_control_device_v1_primary_selection(struct ::ext_data_control_offer_v1 *id) override
     {
+        QMutexLocker locker(&m_primarySelectionLock);
         if (!id) {
             m_receivedPrimarySelection.reset();
         } else {
@@ -468,9 +476,11 @@ protected:
     }
 
 private:
+    QRecursiveMutex m_selectionLock;
     std::unique_ptr<DataControlSource> m_selection; // selection set locally
     std::unique_ptr<DataControlOffer> m_receivedSelection; // latest selection set from externally to here
 
+    QRecursiveMutex m_primarySelectionLock;
     std::unique_ptr<DataControlSource> m_primarySelection; // selection set locally
     std::unique_ptr<DataControlOffer> m_receivedPrimarySelection; // latest selection set from externally to here
     friend WaylandClipboard;
@@ -478,106 +488,92 @@ private:
 
 void DataControlDevice::setSelection(std::unique_ptr<DataControlSource> selection)
 {
-    set_selection(selection->object());
+    {
+        QMutexLocker locker(&m_selectionLock);
+        set_selection(selection->object());
 
-    // Note the previous selection is destroyed after the set_selection request.
-    m_selection = std::move(selection);
-    connect(m_selection.get(), &DataControlSource::cancelled, this, [this]() {
-        m_selection.reset();
-    });
+        // Note the previous selection is destroyed after the set_selection request.
+        m_selection = std::move(selection);
+        connect(m_selection.get(), &DataControlSource::cancelled, this, [this]() {
+            m_selection.reset();
+        });
+    }
 
     Q_EMIT selectionChanged();
 }
 
 void DataControlDevice::setPrimarySelection(std::unique_ptr<DataControlSource> selection)
 {
-    set_primary_selection(selection->object());
+    {
+        QMutexLocker locker(&m_primarySelectionLock);
+        set_primary_selection(selection->object());
 
-    // Note the previous selection is destroyed after the set_primary_selection request.
-    m_primarySelection = std::move(selection);
-    connect(m_primarySelection.get(), &DataControlSource::cancelled, this, [this]() {
-        m_primarySelection.reset();
-    });
+        // Note the previous selection is destroyed after the set_primary_selection request.
+        m_primarySelection = std::move(selection);
+        connect(m_primarySelection.get(), &DataControlSource::cancelled, this, [this]() {
+            m_primarySelection.reset();
+        });
+    }
 
     Q_EMIT primarySelectionChanged();
 }
-class Keyboard;
-// We are binding to Seat/Keyboard manually because we want to react to gaining focus but inside Qt the events are Qt and arrive to late
-class KeyboardFocusWatcher : public QWaylandClientExtensionTemplate<KeyboardFocusWatcher>, public QtWayland::wl_seat
-{
-    Q_OBJECT
-public:
-    KeyboardFocusWatcher()
-        : QWaylandClientExtensionTemplate(5)
-    {
-        initialize();
-        auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
-        auto display = waylandApp->display();
-        // so we get capabilities
-        wl_display_roundtrip(display);
-    }
-    ~KeyboardFocusWatcher() override
-    {
-        if (isActive()) {
-            release();
-        }
-    }
-    void seat_capabilities(uint32_t capabilities) override
-    {
-        const bool hasKeyboard = capabilities & capability_keyboard;
-        if (hasKeyboard && !m_keyboard) {
-            m_keyboard = std::make_unique<Keyboard>(get_keyboard(), *this);
-        } else if (!hasKeyboard && m_keyboard) {
-            m_keyboard.reset();
-        }
-    }
-    bool hasFocus() const
-    {
-        return m_focus;
-    }
-Q_SIGNALS:
-    void keyboardEntered();
 
-private:
-    friend Keyboard;
-    bool m_focus = false;
-    std::unique_ptr<Keyboard> m_keyboard;
-};
-
-class Keyboard : public QtWayland::wl_keyboard
+class ClipboardThread : public QThread
 {
 public:
-    Keyboard(::wl_keyboard *keyboard, KeyboardFocusWatcher &seat)
-        : wl_keyboard(keyboard)
-        , m_seat(seat)
+    ClipboardThread(wl_display *display)
+        : m_display(display)
     {
-    }
-    ~Keyboard()
-    {
-        release();
+        m_queue = wl_display_create_queue_with_name(m_display, "ksystemclipboard queue");
     }
 
-private:
-    void keyboard_enter([[maybe_unused]] uint32_t serial, [[maybe_unused]] wl_surface *surface, [[maybe_unused]] wl_array *keys) override
+    ~ClipboardThread()
     {
-        m_seat.m_focus = true;
-        Q_EMIT m_seat.keyboardEntered();
+        if (m_callback) {
+            wl_callback_destroy(m_callback);
+        }
+        wl_event_queue_destroy(m_queue);
     }
-    void keyboard_leave([[maybe_unused]] uint32_t serial, [[maybe_unused]] wl_surface *surface) override
+
+    void syncQueue()
     {
-        m_seat.m_focus = false;
+        // wl_display_dispatch_queue will block until it gets an event, as KSystemClipboard
+        // has the lifespan of the QApplication and the QPA is still alive
+        // we need something to explicitly wake up the event queue up before the main thread
+        // blocks waiting for this thread.
+        auto wrapped_display = wl_proxy_create_wrapper(m_display);
+        wl_proxy_set_queue(static_cast<wl_proxy *>(wrapped_display), m_queue);
+        m_callback = wl_display_sync(static_cast<struct wl_display *>(wrapped_display));
+        wl_display_flush(m_display);
+        wl_proxy_wrapper_destroy(wrapped_display);
     }
-    KeyboardFocusWatcher &m_seat;
+
+    void run() override
+    {
+        int ret = 0;
+        while (ret >= 0 && !qGuiApp->closingDown()) {
+            ret = wl_display_dispatch_queue(m_display, m_queue);
+        }
+    }
+    wl_callback *m_callback = nullptr;
+    wl_event_queue *m_queue = nullptr;
+    wl_display *m_display = nullptr;
 };
 
 WaylandClipboard::WaylandClipboard(QObject *parent)
     : KSystemClipboard(parent)
-    , m_keyboardFocusWatcher(new KeyboardFocusWatcher)
     , m_manager(new DataControlDeviceManager)
 {
+    auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+    if (!waylandApp) {
+        return;
+    }
+
     connect(m_manager.get(), &DataControlDeviceManager::activeChanged, this, [this]() {
         if (m_manager->isActive()) {
             auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+            m_thread = std::make_unique<ClipboardThread>(waylandApp->display());
+
             if (!waylandApp) {
                 return;
             }
@@ -586,7 +582,9 @@ WaylandClipboard::WaylandClipboard(QObject *parent)
             if (!seat) {
                 return;
             }
+            wl_proxy_set_queue(reinterpret_cast<wl_proxy *>(m_manager->object()), m_thread->m_queue);
             m_device.reset(new DataControlDevice(m_manager->get_data_device(seat)));
+            m_device->moveToThread(m_thread.get());
 
             connect(m_device.get(), &DataControlDevice::receivedSelectionChanged, this, [this]() {
                 // When our source is still valid, so the offer is for setting it or we emit changed when it is cancelled
@@ -607,16 +605,25 @@ WaylandClipboard::WaylandClipboard(QObject *parent)
             connect(m_device.get(), &DataControlDevice::primarySelectionChanged, this, [this]() {
                 Q_EMIT changed(QClipboard::Selection);
             });
+            m_thread->start();
 
         } else {
             m_device.reset();
+            m_thread->wait();
+            m_thread.reset();
         }
     });
 
     m_manager->instantiate();
 }
 
-WaylandClipboard::~WaylandClipboard() = default;
+WaylandClipboard::~WaylandClipboard()
+{
+    if (m_thread->isRunning()) {
+        m_thread->syncQueue();
+        m_thread->wait();
+    }
+}
 
 WaylandClipboard *WaylandClipboard::create(QObject *parent)
 {
@@ -639,23 +646,8 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
         return;
     }
 
-    // roundtrip to have accurate focus state when losing focus but setting mime data before processing wayland events.
-    auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
-    auto display = waylandApp->display();
-    wl_display_roundtrip(display);
-
-    // If the application is focused, use the normal mechanism so a future paste will not deadlock itselfs
-    if (m_keyboardFocusWatcher->hasFocus()) {
-        QGuiApplication::clipboard()->setMimeData(mime, mode);
-        // if we short-circuit the ext_data_device, when we receive the data
-        // we cannot identify ourselves as the owner
-        // because of that we act like it's a synchronous action to not confuse klipper.
-        wl_display_roundtrip(display);
-        return;
-    }
-    // If not, set the clipboard once the app receives focus to avoid the deadlock
-    connect(m_keyboardFocusWatcher.get(), &KeyboardFocusWatcher::keyboardEntered, this, &WaylandClipboard::gainedFocus, Qt::UniqueConnection);
     auto source = std::make_unique<DataControlSource>(m_manager->create_data_source(), mime);
+    source->moveToThread(m_thread.get());
     if (mode == QClipboard::Clipboard) {
         m_device->setSelection(std::move(source));
     } else if (mode == QClipboard::Selection) {
@@ -663,31 +655,18 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
     }
 }
 
-void WaylandClipboard::gainedFocus()
-{
-    disconnect(m_keyboardFocusWatcher.get(), &KeyboardFocusWatcher::keyboardEntered, this, nullptr);
-    // QClipboard takes ownership of the QMimeData so we need to transfer and unset our selections
-    if (auto &selection = m_device->m_selection) {
-        std::unique_ptr<QMimeData> data = selection->releaseMimeData();
-        selection.reset();
-        QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Clipboard);
-    }
-    if (auto &primarySelection = m_device->m_primarySelection) {
-        std::unique_ptr<QMimeData> data = primarySelection->releaseMimeData();
-        primarySelection.reset();
-        QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Selection);
-    }
-}
-
 void WaylandClipboard::clear(QClipboard::Mode mode)
 {
     if (!m_device) {
         return;
     }
+
     if (mode == QClipboard::Clipboard) {
+        QMutexLocker locker(&m_device->m_selectionLock);
         m_device->set_selection(nullptr);
         m_device->m_selection.reset();
     } else if (mode == QClipboard::Selection) {
+        QMutexLocker locker(&m_device->m_primarySelectionLock);
         m_device->set_primary_selection(nullptr);
         m_device->m_primarySelection.reset();
     }
@@ -699,24 +678,42 @@ const QMimeData *WaylandClipboard::mimeData(QClipboard::Mode mode) const
         return nullptr;
     }
 
+    // WaylandClipboard owns the mimedata, but the caller can read it until the next event loop runs.
+    auto lockWithUnlockLater = [this](QRecursiveMutex &mutex) {
+        mutex.lock();
+        QTimer::singleShot(0, this, [&mutex, guard = QPointer<DataControlDevice>(m_device.get())] {
+            if (!guard) {
+                return;
+            }
+            mutex.unlock();
+        });
+    };
+
     // return our locally set selection if it's not cancelled to avoid copying data to ourselves
     if (mode == QClipboard::Clipboard) {
+        QMutexLocker lock(&m_device->m_selectionLock);
         if (m_device->selection()) {
             return m_device->selection();
         }
+
         // This application owns the clipboard via the regular data_device, use it so we don't block ourselves
         if (QGuiApplication::clipboard()->ownsClipboard()) {
             return QGuiApplication::clipboard()->mimeData(mode);
         }
+        lockWithUnlockLater(m_device->m_selectionLock);
         return m_device->receivedSelection();
     } else if (mode == QClipboard::Selection) {
+        QMutexLocker lock(&m_device->m_primarySelectionLock);
         if (m_device->primarySelection()) {
             return m_device->primarySelection();
         }
+
         // This application owns the primary selection via the regular primary_selection_device, use it so we don't block ourselves
         if (QGuiApplication::clipboard()->ownsSelection()) {
             return QGuiApplication::clipboard()->mimeData(mode);
         }
+
+        lockWithUnlockLater(m_device->m_primarySelectionLock);
         return m_device->receivedPrimarySelection();
     }
     return nullptr;
