@@ -158,7 +158,7 @@ private:
     /** reads data from a file descriptor with a timeout of 1 second
      *  true if data is read successfully
      */
-    static bool readData(int fd, QByteArray &data);
+    static bool readData(int fd, QByteArray &data, const QString &mimeType);
     QStringList m_receivedFormats;
     mutable QHash<QString, QVariant> m_data;
 };
@@ -221,7 +221,7 @@ QVariant WlrDataControlOffer::retrieveData(const QString &mimeType, QMetaType ty
     QFile readPipe;
     if (readPipe.open(pipeFds[0], QIODevice::ReadOnly)) {
         QByteArray data;
-        if (readData(pipeFds[0], data)) {
+        if (readData(pipeFds[0], data, mime)) {
             close(pipeFds[0]);
 
             if (mimeType == applicationQtXImageLiteral()) {
@@ -251,7 +251,7 @@ QVariant WlrDataControlOffer::retrieveData(const QString &mimeType, QMetaType ty
     return QVariant();
 }
 
-bool WlrDataControlOffer::readData(int fd, QByteArray &data)
+bool WlrDataControlOffer::readData(int fd, QByteArray &data, const QString &mimeType)
 {
     pollfd pfds[1];
     pfds[0].fd = fd;
@@ -261,18 +261,18 @@ bool WlrDataControlOffer::readData(int fd, QByteArray &data)
         const int ready = poll(pfds, 1, 1000);
         if (ready < 0) {
             if (errno != EINTR) {
-                qWarning("WlrDataControlOffer: poll() failed: %s", strerror(errno));
+                qWarning("WlrDataControlOffer: poll() failed for mimeType %s: %s", qPrintable(mimeType), strerror(errno));
                 return false;
             }
         } else if (ready == 0) {
-            qWarning("WlrDataControlOffer: timeout reading from pipe");
+            qWarning("WlrDataControlOffer: timeout reading from pipe for mimeType %s", qPrintable(mimeType));
             return false;
         } else {
             char buf[4096];
             int n = read(fd, buf, sizeof buf);
 
             if (n < 0) {
-                qWarning("WlrDataControlOffer: read() failed: %s", strerror(errno));
+                qWarning("WlrDataControlOffer: read() failed for mimeType %s: %s", qPrintable(mimeType), strerror(errno));
                 return false;
             } else if (n == 0) {
                 return true;
@@ -322,7 +322,7 @@ WlrDataControlSource::WlrDataControlSource(struct ::zwlr_data_control_source_v1
     for (const QString &format : formats) {
         offer(format);
     }
-    if (mimeData->hasText()) {
+    if (mimeData->hasText() && !formats.contains(utf8Text())) {
         // ensure GTK applications get this mimetype to avoid them discarding the offer
         offer(QStringLiteral("text/plain;charset=utf-8"));
     }
@@ -340,7 +340,7 @@ WlrDataControlSource::WlrDataControlSource(struct ::zwlr_data_control_source_v1
 void WlrDataControlSource::zwlr_data_control_source_v1_send(const QString &mime_type, int32_t fd)
 {
     QString send_mime_type = mime_type;
-    if (send_mime_type == QStringLiteral("text/plain;charset=utf-8")) {
+    if (send_mime_type == utf8Text() && !m_mimeData->hasFormat(utf8Text())) {
         // if we get a request on the fallback mime, send the data from the original mime type
         send_mime_type = QStringLiteral("text/plain");
     }
@@ -478,23 +478,30 @@ private:
 
 void WlrDataControlDevice::setSelection(std::unique_ptr<WlrDataControlSource> selection)
 {
+    set_selection(selection->object());
+
+    // Note the previous selection is destroyed after the set_selection request.
     m_selection = std::move(selection);
     connect(m_selection.get(), &WlrDataControlSource::cancelled, this, [this]() {
         m_selection.reset();
     });
-    set_selection(m_selection->object());
+
     Q_EMIT selectionChanged();
 }
 
 void WlrDataControlDevice::setPrimarySelection(std::unique_ptr<WlrDataControlSource> selection)
 {
+    if (zwlr_data_control_device_v1_get_version(object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
+        set_primary_selection(m_primarySelection->object());
+    }
+
+    // Note the previous selection is destroyed after the set_primary_selection request.
     m_primarySelection = std::move(selection);
     connect(m_primarySelection.get(), &WlrDataControlSource::cancelled, this, [this]() {
         m_primarySelection.reset();
     });
 
     if (zwlr_data_control_device_v1_get_version(object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
-        set_primary_selection(m_primarySelection->object());
         Q_EMIT primarySelectionChanged();
     }
 }
